/*
 * Copyright 2017-2021 Thingstream AG
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef INC_TRACKER_H_
#define INC_TRACKER_H_

#include <stdint.h>
#include <stdbool.h>

#include "nvmem.h"

#if defined(__cplusplus)
extern "C" {
#elif 0
}
#endif

/* Location sources. Add new sources at the end, since the
 * numbers are used as indexes into various tables.
 */

typedef enum LocationType_e
{
    locationGnss,
    locationCell,
    locationWifi,
    locationMeasx,
    locationBLE,

    NUM_LOCATIONS
} LocationType;


/* Invocaton events.
 * Add new events at the end - there are tables that
 * make use of the ordering.
 * Config is not included in this list, since a config event
 * generates a completely different message.
 */
typedef enum InvocationType_e
{
    eventTimer,
    eventButton,
    eventMotion,
    eventPower,

    NUM_EVENTS
} InvocationType;

/**
 * Perform the actions of the tracker app
 * @param invocation one of eventPower, eventMotion, etc
 * @param nvmem the non-volatile store
 * @param modemCreateFlags the flags to use when creating the modem transport
 * @return true on successful send of message, false otherwise
 */
extern bool Tracker_run(InvocationType invocation,
                        NVMem *nvmem,
                        uint16_t modemCreateFlags);


/* upper limit for specifying priority classes */
#define MAX_PRIORITY_CLASS 9

/* upper limit for motion sensitivity setting */
#define MAX_MOTION_SENSITIVITY 5

/* This defines the structure to be stored by nvmem */

struct nvmem_s
{
    /* reserved for use by the platform, perhaps to store
     * version number information, or for flash wear levelling
     */
    uint32_t  platform;

    /* min and max intervals, in minutes */
    struct
    {
        uint16_t  min, max;
    } interval;

    /* The motion detection sensitivity, from 0 (disabled) to
     * 5 (most sensitive) */
    uint8_t motionSensitivity;

    /* for each event location source, the priority class.
     * -1 means never
     *  0 means always
     * Otherwise, priority classes are visited in order
     * 1,2,3...MAX_PRIORITY_CLASS until one group has provided a location.
     */
    int8_t   priority[NUM_LOCATIONS];
};

/* increment this when the layout changes - platforms might
 * choose to incorporate it into the value they store in
 * the platform field.
 */
#define NVMEM_FORMAT_VERSION  2


#if defined(__cplusplus)
}
#endif

#endif /* INC_TRACKER_H_ */
